/* ElGamal using GMP

   fktn.c

   Copyright (C) 1999/2000 Christian Koch
   ckoch@et.htwk-leipzig.de

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; see the file COPYING.  If not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */



#include <windows.h>
#include <math.h>
#include <gmp.h>
#include "global.h"
#include "fktn.h"


void
genkeypair(PubKey *pk, SecKey *sk, int const L)
{
    mpz_t mpzMaske;

    /* Einser-Maske der Lnge L erzeugen */
    mpz_init_set_ui(mpzMaske, 1);
    mpz_mul_2exp(mpzMaske, mpzMaske, L);
    mpz_sub_ui(mpzMaske, mpzMaske, 1);

    /* Bestimmung einer starken Primzahl p der Lnge L Bit */
    srandom(time(NULL));
    do
    {
        mpz_random(pk->p, mpz_size(mpzMaske) + 2);
        mpz_and(pk->p, pk->p, mpzMaske);
        nextstrongprime(pk->p, pk->p);
    }
    while (mpz_sizeinbase(pk->p, 2) != L);
    mpz_set(sk->p, pk->p);

    /* Wahl eines primitiven Elementes alpha bzgl. Zp */
    srandom(time(NULL));
    do
    {
        do
        {
            mpz_random(pk->alpha, mpz_size(pk->p));
            mpz_and(pk->alpha, pk->alpha, mpzMaske);
        }
        while (mpz_cmp(pk->alpha, pk->p) > 0);

        /* p - 1 = [2] * [(p - 1) /2] */
        mpz_powm_ui(pk->beta, pk->alpha, 2, pk->p);
        if (mpz_cmp_ui(pk->beta, 1) == 0)
        {
            continue;
        }
        mpz_sub_ui(pk->beta, pk->p, 1);
        mpz_tdiv_q_2exp(pk->beta, pk->beta, 1);
        mpz_powm(pk->beta, pk->alpha, pk->beta, pk->p);
    }
    while (mpz_cmp_ui(pk->beta, 1) == 0);

    /* Bestimmung eines a mit 1 <= a <= p-2 */
    mpz_sub_ui(pk->beta, pk->p, 2);
    do
    {
        mpz_random(sk->a, mpz_size(pk->p));
        mpz_and(sk->a, sk->a, mpzMaske);
    }
    while (mpz_cmp(sk->a, pk->beta) > 0);
    mpz_clear(mpzMaske);

    /* Bestimmung des beta gem beta = alpha ** a mod p */
    mpz_powm(pk->beta, pk->alpha, sk->a, pk->p);

    return;
}



mpz_t *
encrypt(mpz_t * const mpzIn_p, PubKey const pk, int const no_of_blocks)
{
    int i;
    mpz_t mpztemp, mpzMaske, mpzlambda, *mpzOut_p;

    /* Anforderung von Speicher und Initialisierung der Variablen */
    mpzOut_p = (mpz_t *) VirtualAlloc(NULL, 2 * no_of_blocks * sizeof (mpz_t),\
                                      MEM_COMMIT, PAGE_READWRITE);
    if (mpzOut_p == NULL)
    {
        return (mpzOut_p);
    }

    for (i = 0; i < 2 * no_of_blocks; i++)
    {
        mpz_init(mpzOut_p[i]);
    }

    /* Einser-Maske der Lnge L erzeugen */
    mpz_init_set_ui(mpzMaske, 1);
    mpz_mul_2exp(mpzMaske, mpzMaske, mpz_sizeinbase(pk.p, 2));
    mpz_sub_ui(mpzMaske, mpzMaske, 1);

    /* Verschlsselung der einzelnen Blcke */
    mpz_init(mpztemp);
    mpz_init(mpzlambda);
    srandom(time(NULL));
    for (i = 0; i < no_of_blocks; i++)
    {
        /* Bestimmung eines lambda mit 1 <= lambda <= p-2 */
        mpz_sub_ui(mpztemp, pk.p, 2);
        do
        {
            mpz_random(mpzlambda, mpz_size(pk.p));
            mpz_and(mpzlambda, mpzlambda, mpzMaske);
        }
        while (mpz_cmp(mpzlambda, mpztemp) > 0);

        /* Berechnung nach ElGamal */
        mpz_powm(mpzOut_p[2*i], pk.alpha, mpzlambda, pk.p);
        mpz_powm(mpzOut_p[2*i+1], pk.beta, mpzlambda, pk.p);
        mpz_mul(mpzOut_p[2*i+1], mpzOut_p[2*i+1], mpzIn_p[i]);
        mpz_mod(mpzOut_p[2*i+1], mpzOut_p[2*i+1], pk.p);
    }
    mpz_clear(mpztemp);
    mpz_clear(mpzMaske);
    mpz_clear(mpzlambda);

    return mpzOut_p;
}



mpz_t *
decrypt(mpz_t * const mpzIn_p, SecKey const sk, int const no_of_blocks)
{
    int i;
    mpz_t *mpzOut_p;

    /* Anforderung von Speicher und Initialisierung der Variablen */
    mpzOut_p = (mpz_t *) VirtualAlloc(NULL, no_of_blocks * sizeof (mpz_t),\
                                      MEM_COMMIT, PAGE_READWRITE);
    if (mpzOut_p == NULL)
    {
        return (mpzOut_p);
    }

    for (i = 0; i < no_of_blocks; i++)
    {
        mpz_init(mpzOut_p[i]);
    }

    /* Entschlsselung der einzelnen Blcke */
    for (i = 0; i < no_of_blocks; i++)
    {
        /* Berechnung nach ElGamal */
        mpz_powm(mpzOut_p[i], mpzIn_p[2*i], sk.a, sk.p);
        mpz_invert(mpzOut_p[i], mpzOut_p[i], sk.p);
        mpz_mul(mpzOut_p[i], mpzOut_p[i], mpzIn_p[2*i+1]);
        mpz_mod(mpzOut_p[i], mpzOut_p[i], sk.p);
    }

    return mpzOut_p;
}



void
nextstrongprime(mpz_t mpzPrime, mpz_t const mpzNumber)
{
    mpz_t mpztemp;

    /* ist mpzNumber gerade? */
    mpz_set(mpzPrime, mpzNumber);
    if ((mpz_get_ui(mpzPrime) % 2) == 0)
    {
        mpz_add_ui(mpzPrime, mpzPrime, 1);
        if (mpz_probab_prime_p(mpzPrime, 25) == 1)
        {
            return;
        }
    }

    /* Suche nach einer Primzahl p, so da auch (p-1)/2 prim ist */
    mpz_init(mpztemp);
    do
    {
        mpz_add_ui(mpzPrime, mpzPrime, 2);
        while (mpz_probab_prime_p(mpzPrime, 25) != 1)
        {
            mpz_add_ui(mpzPrime, mpzPrime, 2);
        }
        mpz_sub_ui(mpztemp, mpzPrime, 1);
        mpz_tdiv_q_2exp(mpztemp, mpztemp, 1);
    }
    while (mpz_probab_prime_p(mpztemp, 25) != 1);
    mpz_clear(mpztemp);

    return;
}



mpz_t *
str2arr(int * const no_of_blocks_p, unsigned char const *str_p, int const bits_per_block)
{
    int i, bits_left, str_len;
    mpz_t mpztemp, *mpzarr_p;

    /* Bestimmung der bentigten Anzahl Blcke */
    str_len = lstrlen(str_p);
    *no_of_blocks_p = ceil((double) (str_len * 8) / bits_per_block);

    /* Anforderung von Speicher und Initialisierung der Variablen */
    mpzarr_p = (mpz_t *) VirtualAlloc(NULL, *no_of_blocks_p * sizeof (mpz_t),\
                                      MEM_COMMIT, PAGE_READWRITE);
    if (mpzarr_p == NULL)
    {
        return (mpzarr_p);
    }

    for (i = 0; i < *no_of_blocks_p; i++)
    {
        mpz_init(mpzarr_p[i]);
    }

    /* Aufteilung des Strings in Blcke der Lnge bits_per_block */
    mpz_init(mpztemp);
    for (i = 0; i < str_len; i++)
    {
        bits_left = bits_per_block - (i * 8) % bits_per_block;
        if (bits_left < 8)
        {
            /* Aufteilung des Zeichens auf zwei Blcke */
            mpz_set_ui(mpztemp, str_p[i] >> (8 - bits_left));
            mpz_add(mpzarr_p[(i * 8) / bits_per_block],\
                    mpzarr_p[(i * 8) / bits_per_block], mpztemp);

            mpz_set_ui(mpztemp, ((str_p[i] << bits_left) & 255) >> bits_left);
            mpz_mul_2exp(mpztemp, mpztemp, bits_per_block - (8 - bits_left));
            mpz_add(mpzarr_p[((i * 8) / bits_per_block) + 1],\
                    mpzarr_p[((i * 8) / bits_per_block) + 1], mpztemp);
        }
        else
        {
            /* Abbildung des Zeichen in einen Block */
            mpz_set_ui(mpztemp, str_p[i]);
            mpz_mul_2exp(mpztemp, mpztemp, bits_left - 8);
            mpz_add(mpzarr_p[(i * 8) / bits_per_block],\
                    mpzarr_p[(i * 8) / bits_per_block], mpztemp);
        }
    }
    mpz_clear(mpztemp);

    return mpzarr_p;
}



unsigned char *
arr2str(int const no_of_blocks, mpz_t * const mpzarr_p, int const bits_per_block)
{
    unsigned char *str_p;
    int i, bits_left, str_len;
    mpz_t mpztemp;

    /* Bestimmung der Anzahl der bentigten Zeichen */
    str_len = (no_of_blocks * bits_per_block) / 8;

    /* Anforderung von Speicher */
    str_p = (unsigned char *) VirtualAlloc(NULL, str_len * sizeof (unsigned char) + 1,\
                                           MEM_COMMIT, PAGE_READWRITE);
    if (str_p == NULL)
    {
        return (str_p);
    }

    /* Rekonstruktion der Zeichenkette */
    mpz_init(mpztemp);
    for (i = 0; i < str_len; i++)
    {
        /* Berechnung freier Bits im aktuellen Block */
        bits_left = bits_per_block - (i * 8) % bits_per_block;

        if (bits_left < 8)
        {
            /* Zusammensetzung eines Zeichens aus zwei Blcken */
            str_p[i] = mpz_get_ui(mpzarr_p[(i * 8) / bits_per_block]);
            str_p[i] <<= (8 - bits_left);

            mpz_tdiv_q_2exp(mpztemp, mpzarr_p[((i * 8) / bits_per_block) + 1],\
                            bits_per_block - (8 - bits_left));
            str_p[i] += mpz_get_ui(mpztemp);
        }
        else
        {
            /* Abbildung eines Teils des Blockes auf ein Zeichen */
            mpz_tdiv_q_2exp(mpztemp, mpzarr_p[(i * 8) / bits_per_block], bits_left - 8);
            str_p[i] = mpz_get_ui(mpztemp);
        }
    }
    mpz_clear(mpztemp);

    str_p[str_len] = '\0';
    return str_p;
}
